@@ -1,8 +1,18 @@
+from abc import ABC, abstractmethod
 from collections import defaultdict
-from collections.abc import MutableMapping
 from dataclasses import dataclass
-from itertools import chain
-from typing import TYPE_CHECKING, Dict, Iterable, Optional, Tuple
+from typing import (
+    TYPE_CHECKING,
+    Any,
+    Callable,
+    Dict,
+    Iterable,
+    Iterator,
+    Mapping,
+    MutableMapping,
+    Optional,
+    Tuple,
+)
 
 from dvc_objects.errors import ObjectFormatError
 from pygtrie import ShortKeyError  # noqa: F401, pylint: disable=unused-import
@@ -78,7 +88,76 @@ def _try_load(
     return None
 
 
-class DataIndex(MutableMapping):
+class BaseDataIndex(ABC, Mapping[DataIndexKey, DataIndexEntry]):
+    @abstractmethod
+    def iteritems(
+        self, prefix: Optional[DataIndexKey] = None, shallow: bool = False
+    ) -> Iterator[Tuple[DataIndexKey, DataIndexEntry]]:
+        pass
+
+    @abstractmethod
+    def traverse(self, node_factory: Callable, **kwargs) -> Any:
+        pass
+
+    @abstractmethod
+    def has_node(self, key: DataIndexKey) -> bool:
+        pass
+
+    @abstractmethod
+    def longest_prefix(
+        self, key: DataIndexKey
+    ) -> Tuple[Optional[DataIndexKey], Optional[DataIndexEntry]]:
+        pass
+
+    def ls(self, root_key: DataIndexKey, detail=True):
+        if not detail:
+
+            def node_factory(_, key, children, *args):
+                if key == root_key:
+                    return children
+                else:
+                    return key
+
+        else:
+
+            def node_factory(_, key, children, *args):
+                if key == root_key:
+                    return children
+                else:
+                    return key, self.info(key)
+
+        return self.traverse(node_factory, prefix=root_key)
+
+    def info(self, key: DataIndexKey):
+        try:
+            entry = self[key]
+            assert entry.hash_info
+            isdir = entry.hash_info and entry.hash_info.isdir
+            return {
+                "type": "directory" if isdir else "file",
+                "size": entry.meta.size if entry.meta else 0,
+                "isexec": entry.meta.isexec if entry.meta else False,
+                "isdvc": True,
+                "isout": True,
+                "obj": entry.obj,
+                "entry": entry,
+                entry.hash_info.name: entry.hash_info.value,
+            }
+        except ShortKeyError:
+            return {
+                "type": "directory",
+                "size": 0,
+                "isexec": False,
+                "isdvc": bool(self.longest_prefix(key)),
+                "isout": False,
+                "obj": None,
+                "entry": None,
+            }
+        except KeyError as exc:
+            raise FileNotFoundError from exc
+
+
+class DataIndex(BaseDataIndex, MutableMapping[DataIndexKey, DataIndexEntry]):
     def __init__(self, *args, **kwargs):
         self._trie = Trie(*args, **kwargs)
 
@@ -139,20 +218,24 @@ class DataIndex(MutableMapping):
         for key, entry in self.iteritems(shallow=True, **kwargs):
             self._load(key, entry)
 
-    def has_node(self, key):
+    def has_node(self, key: DataIndexKey) -> bool:
         return self._trie.has_node(key)
 
     def shortest_prefix(self, *args, **kwargs):
         return self._trie.shortest_prefix(*args, **kwargs)
 
-    def longest_prefix(self, *args, **kwargs):
-        return self._trie.longest_prefix(*args, **kwargs)
+    def longest_prefix(
+        self, key: DataIndexKey
+    ) -> Tuple[Optional[DataIndexKey], Optional[DataIndexEntry]]:
+        return self._trie.longest_prefix(key)
 
-    def traverse(self, *args, **kwargs):
+    def traverse(self, *args, **kwargs) -> Any:
         return self._trie.traverse(*args, **kwargs)
 
-    def iteritems(self, prefix=None, shallow=False):
-        kwargs = {"shallow": shallow}
+    def iteritems(
+        self, prefix: Optional[DataIndexKey] = None, shallow: bool = False
+    ) -> Iterator[Tuple[DataIndexKey, DataIndexEntry]]:
+        kwargs: Dict[str, Any] = {"shallow": shallow}
         if prefix:
             kwargs = {"prefix": prefix}
             item = self._trie.longest_prefix(prefix)
@@ -164,32 +247,8 @@ class DataIndex(MutableMapping):
             self._load(key, entry)
             yield key, entry
 
-    def info(self, key):
-        try:
-            entry = self[key]
-            isdir = entry.hash_info and entry.hash_info.isdir
-            return {
-                "type": "directory" if isdir else "file",
-                "size": entry.meta.size if entry.meta else 0,
-                "isexec": entry.meta.isexec if entry.meta else False,
-                "isdvc": True,
-                "isout": True,
-                "obj": entry.obj,
-                "entry": entry,
-                entry.hash_info.name: entry.hash_info.value,
-            }
-        except ShortKeyError:
-            return {
-                "type": "directory",
-                "size": 0,
-                "isexec": False,
-                "isdvc": bool(self.longest_prefix(key)),
-                "isout": False,
-                "obj": None,
-                "entry": None,
-            }
-        except KeyError as exc:
-            raise FileNotFoundError from exc
+    def iterkeys(self, *args, **kwargs):
+        return self._trie.iterkeys(*args, **kwargs)
 
     def _ensure_loaded(self, prefix):
         entry = self._trie.get(prefix)
@@ -203,80 +262,9 @@ class DataIndex(MutableMapping):
             if not entry.obj:
                 raise TreeError
 
-    def ls(self, root_key, detail=True):
+    def ls(self, root_key: DataIndexKey, detail=True):
         self._ensure_loaded(root_key)
-        if not detail:
-
-            def node_factory(_, key, children, *args):
-                if key == root_key:
-                    return children
-                else:
-                    return key
-
-        else:
-
-            def node_factory(_, key, children, *args):
-                if key == root_key:
-                    return children
-                else:
-                    return key, self.info(key)
-
-        return self.traverse(node_factory, prefix=root_key)
-
-
-def _collect_dir(index, prefix, prefix_entry, path, fs, update=False):
-    dir_meta = Meta(nfiles=0, size=0, isdir=True)
-
-    for root, dnames, fnames in fs.walk(path):
-        sub_prefix = fs.path.relparts(root, path) if root != path else ()
-        for name in chain(dnames, fnames):
-            key = (*prefix, *sub_prefix, name)
-            entry_path = fs.path.join(root, name)
-            entry = index.get(key)
-            if entry is None:
-                entry = DataIndexEntry()
-                index[key] = entry
-
-            entry.fs = fs
-            entry.path = entry_path
-            entry.cache = prefix_entry.cache
-            entry.remote = prefix_entry.remote
-
-            # TODO: localfs.walk doesn't currently support detail=True,
-            # so we have to call fs.info() manually
-            meta = Meta.from_info(
-                fs.info(entry_path, refresh=True), fs.protocol
-            )
-            if entry.meta != meta and not update:
-                entry.hash_info = None
-
-            entry.meta = meta
-            dir_meta.nfiles += 1
-            dir_meta.size += meta.size
-
-    return dir_meta
-
-
-def collect(index, path, fs, update=False):
-    # NOTE: converting to list to avoid iterating and modifying the dict the
-    # same time.
-    items = list(index.iteritems(shallow=True))
-    for key, entry in items:
-        entry_path = fs.path.join(path, *key)
-
-        info = fs.info(entry_path, refresh=True)
-
-        fs_meta = Meta.from_info(info, fs.protocol)
-        if entry.meta != fs_meta and not update:
-            entry.hash_info = None
-        entry.meta = fs_meta
-        entry.fs = fs
-        entry.path = entry_path
-
-        if info["type"] == "file":
-            continue
-
-        entry.meta = _collect_dir(index, key, entry, entry_path, fs)
+        return super().ls(root_key, detail=detail)
 
 
 def transfer(index, src, dst):
